#pragma once

#include <concepts>
#include <shared_mutex>

namespace MultiThreading {
    template <typename T>
    class CollisionGuard;
}
template <typename T>
class MultiThreading::CollisionGuard {
private:
    mutable std::shared_mutex mutex;
    T value;
public:

    class Read {
    private:
        std::shared_lock<std::shared_mutex> lock;
        const T* t_ptr;
    public:
        Read(const std::shared_mutex& mtx, const T& val) : t_ptr(&val), lock(mtx) {}
        const T* operator->() const { return t_ptr; }
    };

    class Write {
    private:
        std::unique_lock<std::shared_mutex> lock;
        T* t_ptr;
    public:
        Write(std::shared_mutex& mtx, T& val) : t_ptr(&val), lock(mtx) {}
        T* operator->() { return t_ptr; }
    };
public:
    Read operator->() const { return Read(mutex, value); }
    Write operator->() { return Write(mutex, value); }
public:
    template <typename U = T>
    requires requires(const U& a, const U& b) { a + b; }
    T operator +(const T& rhs) const {
        std::shared_lock lock(mutex);
        return value + rhs;
    }

    template <typename U = T>
    requires requires(U& a, const U& b) { a = b; }
    CollisionGuard& operator =(const T& rhs) {
        std::unique_lock lock(mutex);
        value = rhs;
        return *this;
    }

    explicit operator T() const requires std::copy_constructible<T> {
        std::shared_lock lock(mutex);
        return value;
    }

    template <typename U = T>
    requires requires(U& a, const U& b) { a += b; }
    CollisionGuard& operator +=(const T& rhs) {
        std::unique_lock lock(mutex);
        value += rhs;
        return *this;
    }

    template <typename U = T>
    requires requires(U& a, const U& b) { a -= b; }
    CollisionGuard& operator -=(const T& rhs) {
        std::unique_lock lock(mutex);
        value -= rhs;
        return *this;
    }

    template <typename U = T>
    requires requires(U& a, const U& b) { a *= b; }
    CollisionGuard& operator *=(const T& rhs) {
        std::unique_lock lock(mutex);
        value *= rhs;
        return *this;
    }

    template <typename U = T>
    requires requires(U& a, const U& b) { a /= b; }
    CollisionGuard& operator /=(const T& rhs) {
        std::unique_lock lock(mutex);
        value /= rhs;
        return *this;
    }

    template <typename U = T>
    requires requires(const U& a, const U& b) { a - b; }
    T operator -(const T& rhs) const {
        std::shared_lock lock(mutex);
        return value - rhs;
    }

    template <typename U = T>
    requires requires(const U& a, const U& b) { a * b; }
    T operator *(const T& rhs) const {
        std::shared_lock lock(mutex);
        return value * rhs;
    }

    template <typename U = T>
    requires requires(const U& a, const U& b) { a / b; }
    T operator /(const T& rhs) const {
        std::shared_lock lock(mutex);
        return value / rhs;
    }
public:
    template <typename... Args>
    explicit CollisionGuard(Args&&... args) : value(std::forward<Args>(args)...) {}
};