#pragma once

#include <thread>
#include <MultiThreading/Task.h>

namespace MultiThreading {
    class Thread;
}

class MultiThreading::Thread {
private:
    std::shared_ptr<TaskBase> current_task = nullptr;
    std::atomic<bool> stop = false;
    std::condition_variable cv;
    std::thread worker;
    bool busy = false;
    std::mutex mtx;
private:
    void Runner();
public:
    /// @returns false if the thread is busy.
    /// @param task The task to run.
    /// @note Task is passed by value on purpose so that the original shared_ptr to the task is still good.
    [[nodiscard]] bool SetTask(std::shared_ptr<TaskBase> task);

    /// @returns false if the thread is busy.
    /// @param task The task to run.
    /// @note Task is passed by reference because the frame after next is going to copy it anyway.
    [[nodiscard]] bool SetTask(const std::function<void()>& task);

    /// @returns true if the thread is currently doing work.
    /// @note SetTask will return false if the thread is busy.
    [[nodiscard]] bool Busy() const { return busy; }

    /// Blocks the thread join is called from until execution of this thread exits.
    /// @note Joining this thread from this thread will raise an exception.
    void Join();
public:
    /// Create a thread which will initialize and then wait for a task.
    Thread() { worker = std::thread([this] { this->Runner(); }); }
    /// Crete a thread which will immediately run a task and then wait for another.
    explicit Thread(std::shared_ptr<TaskBase> task);
    /// Crete a thread which will immediately run a task and then wait for another.
    explicit Thread(const std::function<void()>& task);
    /// Waits for the current task to finish (if there is one) and destroys the thread.
    ~Thread();
};