#pragma once

#include <MultiThreading/Thread.h>
#include <vector>
#include <queue>

namespace MultiThreading {
    class ThreadPool;
}

/// A group of threads to run tasks on.
class MultiThreading::ThreadPool {
private:
    std::vector<MultiThreading::Thread*> threads;
    std::queue<std::shared_ptr<TaskBase>> queue;
    std::condition_variable queue_condition;
    std::mutex queue_mutex;
    bool destructing = false;
private:
    /// @returns nullptr if the queue is empty.
    std::shared_ptr<TaskBase> Dequeue();
    void Runner(const std::shared_ptr<TaskBase>& task);
public:
    /// Set a task to be run on the thread-pool.
    /// @param task The task to run.
    bool Enqueue(const std::shared_ptr<MultiThreading::TaskBase>& task);

    /// Set a task to be run on the thread-pool.
    /// @param task The task to run.
    bool Enqueue(const std::function<void()>& task);
public:
    /// @returns The number of threads in the thread pool.
    [[nodiscard]] unsigned int ThreadCount() const { return threads.size(); }

    /// @returns the number of tasks in the queue
    /// @note this excludes the tasks currently being run.
    [[nodiscard]] unsigned int PendingTasks();

    /// @returns Whether a task you enqueue would have to wait for something else to finish before running.
    [[nodiscard]] bool Busy();

    /// Uses a condition variable to wait the calling thread until the queue is empty.
    void WaitQueueEmpty();
public:
    /// Constructs a thread-pool with a given number of threads (or hardware_concurrency as the default)
    /// @param thread_count The number of threads.
    /// @note If you do a *lot* of work on the main thread and hardware_concurrency >= 2, You should use hardware_concurrency -1.
    explicit ThreadPool(unsigned int thread_count = std::thread::hardware_concurrency());

    /// Waits for the threads to empty the queue and destroys the thread pool.
    /// @note This should be one of the very last things your program does before exiting.
    ~ThreadPool();
};