#include <MultiThreading/ThreadPool.h>

using namespace MultiThreading;

ThreadPool::ThreadPool(unsigned int thread_count) {

    for (unsigned int i = 0; i < thread_count; i++)
        threads.push_back(new Thread());
}

bool ThreadPool::Enqueue(const std::shared_ptr<MultiThreading::TaskBase>& task) {
    if (destructing)
        return false;

    std::lock_guard<std::mutex> lock(queue_mutex);

    // Assign it immediately if there's no wait and a thread open.
    if (queue.empty()) {
        for (auto* t: threads) {
            if (t->Busy())
                continue;

            if (!t->SetTask( [this, task] (){ Runner(task); } ))
                throw std::runtime_error("There was a collision while putting the task to the thread.");
            return true;
        }
    }

    // Alternatively it goes in-to the queue.
    queue.push(task);
    return true;
}

std::shared_ptr<TaskBase> ThreadPool::Dequeue() {
    std::lock_guard<std::mutex> lock(queue_mutex);
    if (queue.empty())
        return nullptr;

    auto task = queue.front();
    queue.pop();

    if (queue.empty())
        queue_condition.notify_all();

    return task;
}

bool ThreadPool::Enqueue(const std::function<void()>& task) {
    return Enqueue(Task<void>::Create(task));
}

void ThreadPool::Runner(const std::shared_ptr<TaskBase>& task) {
    auto running_task = task;

    while (running_task) {
        if (!running_task->Complete())
            running_task->Run();

        running_task = Dequeue();
    }
}

unsigned int ThreadPool::PendingTasks() {
    std::lock_guard<std::mutex> lock(queue_mutex);
    return queue.size();
}

ThreadPool::~ThreadPool() {
    destructing = true;
    // Wait for all queued tasks to be running.
    WaitQueueEmpty();

    // delete t waits for the thread to exit gracefully.
    for (auto* t: threads)
        delete t;
}

bool ThreadPool::Busy() {
    if (PendingTasks() != 0)
        return true;

    bool all_busy = true;
    for (auto* t : threads)
        if (!t->Busy()) { all_busy = false; break; }

    return all_busy;
}

void ThreadPool::WaitQueueEmpty() {
    std::unique_lock<std::mutex> lock(queue_mutex);
    queue_condition.wait(lock, [this]{ return queue.empty(); });
}
